/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include <stdarg.h>
#include <signal.h>
#include <fcntl.h>
#include <netinet/tcp.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "httpc.h"
#include "httpc_inet.h"
#include "httpc_client.h"
#include "httpc_util.h"
#include "httpc_handler.h"
#include "httpc_box.h"
#include "httpc_watch.h"
#include "httpc_admin.h"
#include "httpc_session.h"
#include "httpc_db.h"

void reset_privileges(conf_t* conf);

/*
* Adding and removing handlers.
*/

httpc_handler_t* httpcP_add_handler(httpc_t* t) {
  httpc_handler_t* handler = (httpc_handler_t*)malloc(sizeof(httpc_handler_t));
  if (handler) {
    memset(handler, 0, sizeof(*handler));
    handler->httpc = t;
    handler->state = IDLE;
    if (pthread_mutex_lock(&(t->handlers.mutex)) == 0) {
      t->handlers.list =
          (httpc_handler_t**)realloc(t->handlers.list,
          sizeof(httpc_handler_t*) * (t->handlers.count+1));
      if (!t->handlers.list) {
        free(handler);
        return NULL;
      }
      t->handlers.list[(t->handlers.count)++] = handler;
      pthread_mutex_unlock(&(t->handlers.mutex));
    } else {
      free(handler);
      return NULL;
    }
  }
  return handler;
}

void httpcP_remove_handler(httpc_t* t, httpc_handler_t* h) {
  if (pthread_mutex_lock(&(t->handlers.mutex)) == 0) {
    int i;
    for (i=0; i<t->handlers.count; i++) {
      if (t->handlers.list[i] == h) {
        if (i < t->handlers.count-1) {
          memmove(&(t->handlers.list[i]),
                  &(t->handlers.list[i+1]),
                  sizeof(httpc_handler_t*) * (t->handlers.count - (i+1)));
        }
        t->handlers.count--;
        free(h);
        break;
      }
    }
    pthread_mutex_unlock(&(t->handlers.mutex));
  }
}

/*
* Adding and removing clients.
*/

int httpcP_reject(httpc_server_t* s, struct in_addr a) {
  int i; for (i=0; i<s->rejects.count; i++) {
    if (!memcmp(&(s->rejects.list[i]), &a, sizeof(struct in_addr))) {
      fprintf(stderr, "%s: Rejected.\n", inet_ntoa(a));
      return 1;
    }
  }
  return 0;
}

client_t* httpc_add_client(httpc_server_t* s, int fd, struct in_addr a) {
  if (!httpcP_reject(s, a)) {
    client_t* result = (client_t*)malloc(sizeof(client_t));
    memset(result, 0, sizeof(client_t));
    if (pthread_mutex_lock(&(s->clients.mutex)) == 0) {
      s->clients.list = (client_t**)realloc(s->clients.list,
                        (s->clients.count+1) * sizeof(client_t*));
      s->clients.list[s->clients.count++] = result;
      pthread_mutex_unlock(&(s->clients.mutex));
    } else {
      free(result);
      return NULL;
    }
    result->fd     = fd;
    result->server = s;
    memcpy(&(result->address), &a, sizeof(struct in_addr));
//  fprintf(stderr, "New client at %s.\n", inet_ntoa(result->address));
    return result;
  }
  return NULL;
}

void httpc_remove_client_safe(httpc_server_t* s, client_t* client) {
  int i;
  for (i=0; i<s->clients.count; i++) {
    if (s->clients.list[i] == client) {
      if (i < s->clients.count-1) {
        memmove(&(s->clients.list[i]),
                &(s->clients.list[i+1]),
                sizeof(client_t*) * (s->clients.count - (i+1)));
      }
      s->clients.count--;
#ifdef _DEBUG
    fprintf(stderr, "-%s:%u %d %p gone (%d).\n",
            inet_ntoa(client->address), client->port, client->fd,
            client->handler, s->clients.count);
#else
    fprintf(stderr, "-%s: gone.\n", inet_ntoa(client->address));
#endif
      if (client->fd >= 0) close(client->fd);
      free(client);
      break;
    }
  }
}

void httpc_remove_client(httpc_server_t* s, client_t* client) {
  if (pthread_mutex_lock(&(s->clients.mutex)) == 0) {
    httpc_remove_client_safe(s, client);
    pthread_mutex_unlock(&(s->clients.mutex));
  }
}

void httpc_close_client_safe(client_t* c) {
  if (c->fd >= 0) {
    close(c->fd);
    c->fd = -1;
    c->readstate = SILENT;
    c->writestate = SILENT;
    if (c->session) {
      httpc_session_dec(c->server->httpc, c->session);
    }
  }
}

void httpc_close_client(client_t* c) {
  if (pthread_mutex_lock(&(c->server->clients.mutex)) == 0) {
    httpc_close_client_safe(c);
    pthread_mutex_unlock(&(c->server->clients.mutex));
  }
}

/*
* This function is called by a handler thread; it blocks until there is
* a client readable or writable.
*/

client_t* httpcP_client_pop(httpc_handler_t* handler) {
  httpc_t* t = handler->httpc;
  for (;;) {
    int i;
    for (i=0; i<t->servers.count; i++) {
      int j;
      if (pthread_mutex_lock(&(t->servers.list[i]->clients.mutex)) == 0) {
        for (j=0; j<t->servers.list[i]->clients.count; j++) {
          client_t* client = t->servers.list[i]->clients.list[j];
          if (client->readstate == SELECTED) {
            client->readstate = HANDLED;
            client->handler = handler;
            handler->timestamp = time(NULL);
            pthread_mutex_unlock(&(t->servers.list[i]->clients.mutex));
            return client;
          } else if (client->writestate == SELECTED) {
            client->writestate = HANDLED;
            client->handler = handler;
            handler->timestamp = time(NULL);
            pthread_mutex_unlock(&(t->servers.list[i]->clients.mutex));
            return client;
          }
        }
        pthread_mutex_unlock(&(t->servers.list[i]->clients.mutex));
      }
    }
    pthread_cond_wait(&(t->handlers.condition), &(t->handlers.condmutex));
    if (handler->httpc->state == APPLICATION_END ||
        handler->state == DYING) {
      return NULL;
    }
  }
}

/*
* The handler code lives in a separate thread that removes its own
* data abstraction from the list when done looping endlessly.
*/

void* httpcP_thread_handler(void* arg) {
  httpc_handler_t* handler = (httpc_handler_t*)arg;
  httpc_t* t = handler->httpc;
  while (handler->state != DYING) {
    client_t* client = NULL;
    if (pthread_mutex_lock(&(t->handlers.condmutex)) == 0) {
      client = httpcP_client_pop(handler);
      pthread_mutex_unlock(&(t->handlers.condmutex));
    }
    if (client) {
      int sig = HTTPC_SIGNEW;
      handler->state = BUSY;
      if (client->readstate == HANDLED) {
        httpc_handle_http_request(client);
        client->handler = NULL;
      } else if (client->writestate == HANDLED) {
        httpc_handle_write(client);
        client->handler = NULL;
      }
      write(handler->httpc->pipe[1], &sig, sizeof(int));
      handler->state = IDLE;
      t->waitcount--;
    }
  }
  httpcP_remove_handler(t, handler);
  return NULL;
}

/*
* Helper function to the load-balancing httpc_examine_handlers(...)
* Lives in the main thread, spawning new handler-threads.
*/

int httpc_spawn_handler(httpc_t* t) {
  httpc_handler_t* handler = httpcP_add_handler(t);
  if (handler) {
    pthread_t thread;
    if (!pthread_create(&thread, NULL, httpcP_thread_handler, (void*)handler)) {
      return 0;
    }
  }
  return -1;
}

#ifdef _DEBUG1
char* shandlerstate[] = {
  "idle",
  "busy",
  "dying"
};
#endif

int httpc_examine_handlers(httpc_t* t, int treshold) {
  int killed = 0, idle = 0;
#ifdef _DEBUG1
  int statecount[] = { 0, 0, 0 };
  fprintf(stderr, "* %d handlers ", t->handlers.count);
  fflush(stderr);
#endif
  if (pthread_mutex_lock(&(t->handlers.mutex)) == 0) {
    int i;
    for (i=0; i<t->handlers.count; i++) {
      httpc_handler_t* handler = t->handlers.list[i];
#ifdef _DEBUG1
      statecount[handler->state]++;
#endif
      if (handler->state == IDLE) ++idle;
      if (idle >= treshold) {
        handler->state = DYING;
        ++killed;
      }
    }
    pthread_mutex_unlock(&(t->handlers.mutex));
  }
#ifdef _DEBUG1
  fprintf(stderr, "%d idle %d busy %d dying.\n",
                  statecount[0], statecount[1], statecount[2]);
  fflush(stderr);
#endif
  if (killed) {
    pthread_cond_broadcast(&(t->handlers.condition));
    fprintf(stderr, "* Removed %d handlers.\n", killed);
    fflush(stderr);
  }
#ifdef _DEBUG1
  if (t->servers.list[0]->clients.count > 20 && statecount[2] == 0) {
    int sig = HTTPC_SIGNEW;
    fprintf(stderr, "Suspicous silence %d %p.\n",
                    t->servers.list[0]->clients.list[0]->fd,
                    t->servers.list[0]->clients.list[0]->handler);
    write(t->pipe[1], &sig, sizeof(int));
  }
#endif
  return 0;
}

/* stolen from apache */

#ifndef MAX_SECS_TO_LINGER
#define MAX_SECS_TO_LINGER 30
#endif

static void sock_enable_linger(int s)
{
    struct linger li;

    li.l_onoff = 1;
    li.l_linger = MAX_SECS_TO_LINGER;

    if (setsockopt(s, SOL_SOCKET, SO_LINGER,
                   (char *) &li, sizeof(struct linger)) < 0) {
      fprintf(stderr, "setsockopt SO_LINGER.\n");
    }
}

void httpcP_setsockopt(int s) {
  int just_say_no = 1;
  // #include <sys/time.h>
  // struct timeval timeout = { 0, 250000 };
  if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *) &just_say_no,
                 sizeof(int)) < 0) {
    perror("setsockopt IPPROTO_TCP");
  }
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &just_say_no,
		sizeof(int)) < 0) {
    perror("setsockopt SO_REUSEADDR");
  }
  sock_enable_linger(s);
}

/*
* Code that maybe should not be here; it adds to a list of client addresses
* that should not be accepted.
*/

void httpc_add_rejects(httpc_server_t* server, int root) {
  int n = c_first_child_named(server->httpc->config, root, "reject");
  while (n != -1) {
    struct in_addr address;
    if (resolve(c_value(server->httpc->config, n), &address) == 0) {
      server->rejects.list = (struct in_addr*)realloc(server->rejects.list,
        sizeof(struct in_addr) * (server->rejects.count+1));
      if (!server->rejects.list) {
        server->rejects.count = 0;
        return;
      }
      memcpy(&(server->rejects.list[server->rejects.count++]),
             &address,
             sizeof(struct in_addr));
    }
    n = c_next_brother_named(server->httpc->config, n, "reject");
  }
}

/*
* Adds a service on host:port
*/
#if 0
void getaddr() {
     getaddrinfo(const char *nodename, const char *servname,
             const struct addrinfo *hints, struct addrinfo **res);

     struct addrinfo {                                                  *
          int     ai_flags;     /* AI_PASSIVE, AI_CANONNAME, AI_NUMERICHOST */
          int     ai_family;    /* PF_xxx */
          int     ai_socktype;  /* SOCK_xxx */
          int     ai_protocol;  /* 0 or IPPROTO_xxx for IPv4 and IPv6 */
          size_t  ai_addrlen;   /* length of ai_addr */
          char   *ai_canonname; /* canonical name for nodename */
          struct sockaddr  *ai_addr; /* binary address */
          struct addrinfo  *ai_next; /* next structure in linked list */
     };
}
#endif
#undef USE_ADDRINFO

int httpcP_add_server(httpc_t* t, char* host, short port, int root) {
  httpc_server_t* server;
#ifdef USE_ADDRINFO
  struct addrinfo * ai, * aip = NULL, hints = {
	AI_PASSIVE | AI_NUMERICHOST, PF_INET, SOCK_STREAM
  };
#endif

  int sockfd;
#ifndef USE_ADDRINFO
  struct sockaddr_in sin;
  memset(&sin, 0, sizeof(sin));
  if (resolve(host, &(sin.sin_addr))) {
    perror("resolve:");
    return -1;
  }
#else
  if (getaddrinfo(host, NULL, &hints, &ai)) {
    perror("getaddrinfo:");
    return -1;
  }
  aip = ai;
#endif

#ifndef USE_ADDRINFO
  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
#else
  if ((sockfd = socket(
     aip->ai_family, aip->ai_socktype, aip->ai_protocol)) == -1) {
#endif
    perror("socket:");
    return -1;
  }

  httpcP_setsockopt(sockfd);
  fcntl(sockfd, F_SETFL, O_NONBLOCK);

#ifndef USE_ADDRINFO
  sin.sin_family = AF_INET;
  sin.sin_port   = htons(port);
  if (bind(sockfd, (struct sockaddr*)&sin, sizeof(sin))) {
#else
  if (bind(sockfd, aip->ai_addr, aip->ai_addrlen)) {
#endif
    perror("bind:");
    return -1;
  }
  if (listen(sockfd, 10)) {
    perror("listen:");
    return -1;
  }
  t->servers.list = 
    (httpc_server_t**)realloc(t->servers.list,
    (t->servers.count+1) * sizeof(httpc_server_t*));
  server = (httpc_server_t*)malloc(sizeof(httpc_server_t));
  t->servers.list[(t->servers.count)++] = server;
  memset(server, 0, sizeof(*server));
  server->httpc          = t;
  // server->ipaddress      = sin.sin_addr;
  server->port           = port;
  server->sockfd         = sockfd;
  server->serverrootnode = root;
  httpc_add_rejects(server, root);
  httpc_add_serverinfo(server, root);
  pthread_mutex_init(&(server->clients.mutex), NULL);
  fprintf(stderr, "Server:successfully bound to %s:%d.\n", host, port);
  return 0;
}

/*
* A quick loop through all the clients to see who can be removed.
*/

void httpcP_destroy_clients(httpc_t* t) {
  int i;
  for (i=0; i<t->servers.count; i++) {
    if (pthread_mutex_lock(&(t->servers.list[i]->clients.mutex)) == 0) {
      int j;
      for (j=0; j<t->servers.list[i]->clients.count; j++) {
        client_t* client = t->servers.list[i]->clients.list[j];
        if (client->fd < 0 && !client->handler) {
          httpc_remove_client_safe(t->servers.list[i], client);
        }
      }
      pthread_mutex_unlock(&(t->servers.list[i]->clients.mutex));
    }
  }
}

/*
* This is where the main threads finds its end in a never ending loop.
* Here, it brings up all services and starts accepting connections.
*/

int httpcP_startserver(httpc_t* t) {
  int err, i = 0, cnt = 0;
  int node;
  if ((node = c_first_child_named(t->config, -1, "server")) != -1) {
    do {
      int n = c_first_child_named(t->config, node, "host");
      int p = c_first_child_named(t->config, node, "port");
      char* host = c_value(t->config, n);
      char* port = c_value(t->config, p);
      if (n != -1 && p != -1 && host && port) {
        if ((err = httpcP_add_server(t, host, atoi(port), node))) {
          fprintf(stderr, "ERROR:httpc_startserver %d: %s:%s\n", i, host, port);
          cnt++;
        } else {
          fprintf(stderr, "httpc_startserver %d: %s:%s\n", i, host, port);
        }
      }
      ++i;
    } while ((node = c_next_brother_named(t->config, node, "server")) != -1);
  }
  reset_privileges(t->config);
  if (cnt == i) {
    fprintf(stderr, "All services failed.  Exit.\n");
    return -1;
  }
  t->state = APPLICATION_RUNNING;
  // the watcher thread
  if (httpc_watcherthread(t)) {
    fprintf(stderr, "Could not spawn watcher thread.  Exit.\n");
    return -1;
  }
  // the loop
  while (t->state != APPLICATION_END) {
    int cnt, fdmax = 0;
    fd_set readset, writeset;
    FD_ZERO(&readset);
    FD_ZERO(&writeset);
    if (t->pipe[0] > fdmax) fdmax = t->pipe[0];
    FD_SET(t->pipe[0], &readset);
    if (t->adminsockfd > fdmax) fdmax = t->adminsockfd;
    if (t->adminsockfd >= 0) FD_SET(t->adminsockfd, &readset);
    for (i=0; i<t->servers.count; i++) {
      int fd = t->servers.list[i]->sockfd;
      if (fd > fdmax) fdmax = fd;
      FD_SET(fd, &readset);
      if (pthread_mutex_lock(&(t->servers.list[i]->clients.mutex)) == 0) {
        int j;
        for (j=0; j<t->servers.list[i]->clients.count; j++) {
          client_t* client = t->servers.list[i]->clients.list[j];
          if ((fd = t->servers.list[i]->clients.list[j]->fd) >= 0) {
            if (client->readstate == SELECTABLE) {
              if (fd > fdmax) fdmax = fd;
              FD_SET(fd, &readset);
            }
            if (client->writestate == SELECTABLE) {
              fd = t->servers.list[i]->clients.list[j]->fd;
              if (fd > fdmax) fdmax = fd;
              FD_SET(fd, &writeset);
            }
          }
        }
        pthread_mutex_unlock(&(t->servers.list[i]->clients.mutex));
      }
    }
    if ((cnt = select(fdmax + 1, &readset, &writeset, NULL, NULL)) > 0) {
      if (FD_ISSET(t->pipe[0], &readset)) {
        int sig;
        read(t->pipe[0], &sig, sizeof(int));
        if (sig == HTTPC_SIGSTOP) {
          t->state = APPLICATION_END;
        }
      }
      if (t->adminsockfd >= 0 && FD_ISSET(t->adminsockfd, &readset)) {
        httpc_admin_accept(t);
      }
      for (i=0; i<t->servers.count; i++) {
        if (pthread_mutex_lock(&(t->servers.list[i]->clients.mutex)) == 0) {
          int j;
          for (j=0; j<t->servers.list[i]->clients.count; j++) {
            client_t* client = t->servers.list[i]->clients.list[j];
            if (client->fd >= 0) {
              if (FD_ISSET(client->fd, &readset)) {
                client->readstate = SELECTED;
                pthread_cond_signal(&(t->handlers.condition));
                httpc_touch(t);
                t->waitcount++;
              }
              if (FD_ISSET(client->fd, &writeset)) {
                client->writestate = SELECTED;
                pthread_cond_signal(&(t->handlers.condition));
                t->waitcount++;
              }
            }
          }
          pthread_mutex_unlock(&(t->servers.list[i]->clients.mutex));
        }
        if (FD_ISSET(t->servers.list[i]->sockfd, &readset)) {
          struct sockaddr_in sin;
          int fd, len = sizeof(sin);
          if ((fd = accept(t->servers.list[i]->sockfd,
                           (struct sockaddr*)&sin, &len)) >= 0) {
            client_t* client;
            if ((client =
                 httpc_add_client(t->servers.list[i], fd, sin.sin_addr))) {
#ifdef _DEBUG
              client->port = ntohs(sin.sin_port);
#endif
              client->readstate = SELECTABLE;
            } else {
              close(fd);
            }
          }
        }
      }
    }
    httpcP_destroy_clients(t);
  }
  fprintf(stderr, "Application ends.\n");
  httpc_examine_handlers(t, 0);
  sleep(3);
  for (i=0; i<t->servers.count; i++) {
    if (shutdown(t->servers.list[i]->sockfd, SHUT_RDWR)) {
      perror("shutdown");
    }
    if (pthread_mutex_lock(&(t->servers.list[i]->clients.mutex)) == 0) {
      while (t->servers.list[i]->clients.count) {
        httpc_remove_client_safe(t->servers.list[i],
                                 t->servers.list[i]->clients.list[0]);
      }
      pthread_mutex_unlock(&(t->servers.list[i]->clients.mutex));
    }
    close(t->servers.list[i]->sockfd);
    httpc_db_free(t->servers.list[i]);
  }
  fprintf(stderr, "---- Done.\n");
  return 0;
}

#define GETENV   0
#define LITTERAL 1

char* httpcP_createpath(int how, char* env,
                        char* file, char* buffer, int size) {
  char* home = ((how == GETENV) ? getenv(env) : env);
  if (home) {
    strncpy(buffer, home, size);
    if (home[strlen(home)-1] != '/') {
      strncat(buffer, "/", size);
    }
    strncat(buffer, file, size);
    return buffer;
  }
  return NULL;
}

/*
* Three functions that take care of runtime protection of the process.
* fd[2] is being redirected to a log file,
* the process' privileges are being reset,
* the process goes into the background.
*/

void redirect_logging(conf_t* conf) {
  int nlog = c_first_child_named(conf, -1, "log");
  if (nlog != -1) {
    char* fnlog = c_value(conf, nlog);
    if (fnlog) {
      int fderr = open(fnlog, O_WRONLY|O_APPEND|O_CREAT, 0644);
      if (fderr >= 0) {
        dup2(fderr, fileno(stderr));
        close(fderr);
      }
    }
  }
}

#include <pwd.h>
#include <grp.h>

void reset_privileges(conf_t* conf) {
  int nuid = c_first_child_named(conf, -1, "user");
  int ngid = c_first_child_named(conf, -1, "group");
  if (ngid != -1) {
    struct group* grp = getgrnam(c_value(conf, ngid));
    if (grp) {
      setgid(grp->gr_gid);
    }
  }
  if (nuid != -1 && getuid() == 0) {
    struct passwd* pwd = getpwnam(c_value(conf, nuid));
    if (pwd) {
      setuid(pwd->pw_uid);
    }
  }
  if (getuid() == 0) {
    printf("Refuse to play as root.\n");
    exit(-1);
  }
}

void retreat(conf_t* conf) {
  if (c_first_child_named(conf, -1, "nodaemon") == -1) {
    daemon(1, 1);
  }
}

/*
* Writes a default config file in the current working directory
* and tells the user about it.
*/

void httpcP_write_default_config() {
  int fd = open("httpcd.conf", O_WRONLY|O_CREAT, 0644);
  if (fd >= 0) {
    char defaultconfig[] =
      "server {\n"
      "## Use IP address of interface or name here.\n"
      "  host = \"localhost\"\n"
      "  port = 8088\n"
      "  log = \"/var/tmp/box.log\"\n"
      "  state = \"/var/tmp/box.state\"\n"
      "  maxlines = 20\n"
      "\n"
      "## Paths to html template files that will be pre- and postpended.\n"
      "## (These values default to an empty string)\n"
      "# head = /home/chatbox/files/head.htmlt\n"
      "# bodyprefix = /home/chatbox/files/bodyprefix.htmlt\n"
      "# bodypostfix = /home/chatbox/files/bodypostfix.htmlt\n"
      "\n"
      "## Paths to html template files that will be used for chatlines.\n"
      "## (These values default to strings including '${user}' and '${line}')\n"
      "# line = /home/chatbox/files/line.htmlt\n"
      "# personal = /home/chatbox/files/personal.htmlt\n"
      "# message = /home/chatbox/files/message.htmlt\n"
      "\n"
      "## An item in boxnamespace that's special (define with simple regex).\n"
      "# box = \"[Bb]ox\" {\n"
      "#    name = box\n"
      "## items that can be defined here are:\n"
      "## 'head' 'bodyprefix' 'bodypostfix' 'line' 'personal' 'message'\n"
      "# }\n"
      "\n"
      "## This is one way to obtain user information\n"
      "# authentication = /home/chatbox/auth/boxroot.auth\n"
      "\n"
      "## These are mutually exclusive\n"
      "# behaviour = must_have_cookie\n"
      "# behaviour = use_foreign_cookies\n"
      "}\n"
      "\n"
      "## Unset (comment out, remove) to dissociate from tty.\n"
      "nodaemon\n"
      "\n"
      "## User/group will only be triggered when running as root.\n"
      "## N.B. the program will refuse to run as root.\n"
      "#user = nobody\n"
      "#group = nobody\n"
      "\n"
      "## Log file of all not chatline related info.\n"
      "log = \"/var/tmp/httpcd.log\"\n"
      "\n"
      "## Max. number of threads to be spawned when things get busy.\n"
      "maxthreads = 20\n"
      "minthreads = 4\n"
      "\n"
      "## UNIX socket can be used to control the chatbox by another process.\n"
      "#admin = \"/tmp/.httpcd_admin_socket\"\n";
    write(fd, defaultconfig, strlen(defaultconfig));
    close(fd);
    fprintf(stderr, "Created new config file 'httpcd.conf'.\n"
                    "Please adjust and move to '/etc' or $HTTPCD_HOME.\n");
  } else {
    fprintf(stderr, "Could not open or write configfile 'httpcd.conf'.\n");
  }
}

/*
* This is where we initialize a httpc structure; read a config file
* and tune the process a bit using the above three functions.
*/

httpc_t* httpc_init(char* path) {
  time_t ti = time(NULL);
  struct tm* tm = localtime(&ti);
  httpc_t* t = (httpc_t*)malloc(sizeof(httpc_t));
  memset(t, 0, sizeof(*t));
  t->state = APPLICATION_START;
  if (!path) {
    char cpath[FILENAME_MAX],
         cpdefault[] = "httpcd.conf";
    if (!(path = httpcP_createpath(GETENV, "HTTPCDHOME",
                                   cpdefault, cpath, FILENAME_MAX))) {
      if (!(path = httpcP_createpath(GETENV, "HTTPCD_HOME",
                                     cpdefault, cpath, FILENAME_MAX))) {
        if (!(path = httpcP_createpath(LITTERAL, "/etc",
                                       cpdefault, cpath, FILENAME_MAX))) {
          path = cpdefault;
        }
      }
    }
  }
  t->config = c_swallow(path);
  if (!t->config) {
    free(t);
    fprintf(stderr, "Config file at '%s' is corrupt.\n", path);
    return NULL;
  }
  if (c_first_child(t->config, -1) == -1) {
    httpcP_write_default_config();
    free(t);
    return NULL;
  }
  pthread_mutex_init(&(t->handlers.mutex), NULL);
  pthread_mutex_init(&(t->handlers.condmutex), NULL);
  pthread_cond_init(&(t->handlers.condition), NULL);
  pthread_mutex_init(&(t->sessions.mutex), NULL);
  redirect_logging(t->config);
  retreat(t->config);
  fprintf(stderr, "\n---- %d-%.2d-%.2d %.2d:%.2d:%.2d\n"
                  "httpcd init, using config file '%s'.\n"
                  , tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday
                  , tm->tm_hour, tm->tm_min, tm->tm_sec
                  , path);
  httpc_admin_init(t);
  pipe(t->pipe);
  signal(SIGPIPE, SIG_IGN);
  return t;
}

/*
* This one blocks.
*/

int httpc_start(httpc_t* t) {
  if (httpcP_startserver(t)) {
    return -1;
  }
  return 0;
}

/*
* This one ends it all.
*/

void httpc_stop(httpc_t* t) {
  int sig = HTTPC_SIGSTOP;
  write(t->pipe[1], &sig, sizeof(int));
}
