/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc_format.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include <unistd.h>
#include <fcntl.h>

#include "httpc_db.h"

struct keydata {
  char   name[BOXNAMESIZE];
  time_t time;
};

#ifdef _DEBUG
void printkey(const DBT* key) {
  int i;
  for (i=0; i<key->size; i++) {
    unsigned char* data = (unsigned char*)key->data;
    fprintf(stderr, "%.3d ", data[i]);
  }
  fprintf(stderr, "\n");
}
#endif

/*
* Does a reverse not-bytewise comparison for the second part of the key
* (the timestamp) so that the latest timestamps come first, and they're
* always compared in host byte-order.
*/

int httpcP_db_compare(const DBT *key1, const DBT *key2) {
  struct keydata* key1data = (struct keydata*)(key1->data);
  struct keydata* key2data = (struct keydata*)(key2->data);
  int c;
  if ((c = memcmp(key1data->name,
                  key2data->name,
                  BOXNAMESIZE))) {
    return c;
  } else {
    if (key1->size == BOXNAMESIZE || key2->size == BOXNAMESIZE) return 0;
    return (key2data->time - key1data->time);
  }
}

/*
* Inits a db and a mutex used to ensure singular access to it.
*/

int httpc_db_init(httpc_server_t* server, char* file) {
  BTREEINFO info;
  memset(&info, 0, sizeof(info));
  info.flags = R_DUP;
//  info.cachesize
//  info.maxkeypage
//  info.minkeypage
//  info.psize
  info.compare = httpcP_db_compare;
//  info.prefix
//  info.lorder
  pthread_mutex_init(&(server->boxmutex), NULL);
  if ((server->db = dbopen(file, O_RDWR|O_CREAT, 0662, DB_BTREE, &info))) {
    return 0;
  } else {
    return 1;
  }
}

/*
* The db will be closed after this; for the sake of keeping all the data
* intact, this function _must_ be called.
*/

void httpc_db_free(httpc_server_t* server) {
  DB* db = server->db;
  db->close(db);
  pthread_mutex_destroy(&(server->boxmutex));
}

/*
* Once in every so many inserts (specify hereafter), the db will be examined
* to get rid of records that are too old or too many.
* This will only keep the db (relatively) small in a statistical way;
* this code does _not_ensure_ db smallness, and if for some reason the
* every so manieth insert always happens in the same box, all other boxes
* can grow unlimitedly.  But if anybody knows a better way, contribute !
*/

#define EXAMINE_FREQUENCY 20
int counter = 0;

void httpcP_db_examine(time_t time, char* name,
                       DB* db, DBT* key,
                       int numberlimit, time_t timelimit) {
  DBT value;
  if (db->seq(db, key, &value, R_CURSOR) == 0) {
    int cnt = 1;
    while (db->seq(db, key, &value, R_NEXT) == 0) {
      if (strcmp(name, ((struct keydata*)(key->data))->name)) break;
      if (cnt++ >= numberlimit ||
          ((struct keydata*)(key->data))->time < time - timelimit) {
        db->del(db, key, R_CURSOR);
      }
    }
  }
}

/*
* Inserts a string into the db using the client's box name and the current time 
* as key.
*/

int httpc_db_insert(client_t* client, char* line) {
  httpc_server_t* server = client->server;
  DB* db = server->db;
  DBT key, value;
  struct keydata keydata;
  strncpy(keydata.name, client->session->boxname, BOXNAMESIZE);
  keydata.time = time(NULL);
  key.data = (void*)&keydata;
  key.size = sizeof(struct keydata);
  value.data = line;
  value.size = strlen(line)+1;
  if (pthread_mutex_lock(&(server->boxmutex)) == 0) {
    if (db->put(db, &key, &value, 0)) {
      perror("httpc_db_insert");
    }
    if (++counter == EXAMINE_FREQUENCY) {
// these numbers are arbitrary for now
      httpcP_db_examine(keydata.time, keydata.name, db, &key, 100, 30 * 86400);
      counter = 0;
    }
    pthread_mutex_unlock(&(server->boxmutex));
  } else return ~0;
  return 0;
}

/*
* Duplicates up until a maximum number of lines (which are in turn stringified
* hashes) from the db into an array of strings.
*/

int httpc_db_dup_lines(client_t* client, char** array, int size) {
  if (size > 0) {
    int cnt = 0;
    struct keydata keydata;
    httpc_server_t* server = client->server;
    DB* db = server->db;
    DBT key, value;
    strncpy(keydata.name, client->session->boxname, BOXNAMESIZE);
    keydata.time = time(NULL);
    key.data = (void*)&keydata;
    key.size = sizeof(struct keydata);
    if (pthread_mutex_lock(&(server->boxmutex)) == 0) {
//      memset(&value, 0, sizeof(value));
      if (db->seq(db, &key, &value, R_CURSOR) == 0) {
        char* boxname = client->session->boxname;
        if (!strncmp(key.data, boxname, BOXNAMESIZE)) {
          struct keydata* k = (struct keydata*)key.data;
          if (size > 1 && k->time > time(NULL) + 180) {
            array[cnt++] = strdup("line=\"Silence.\"");
          }
          array[cnt++] = strdup(value.data);
        } else {
          pthread_mutex_unlock(&(server->boxmutex));
          return 0;
        }
//        memset(&value, 0, sizeof(value));
        while (db->seq(db, &key, &value, R_NEXT) == 0 && cnt < size) {
          if (!strncmp(key.data, boxname, BOXNAMESIZE)) {
            array[cnt++] = strdup(value.data);
          } else break;
//          memset(&value, 0, sizeof(value));
        }
      }
      pthread_mutex_unlock(&(server->boxmutex));
    } else return 0;
    return cnt;
  }
  return 0;
}
