/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc_watch.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include "httpc_watch.h"
#include "httpc_client.h"
#include "httpc_session.h"

// don't believe the hype; three connections per second _is_ busy.

static inline int is_high(int visits, int seconds) {
  return ((visits/seconds > 3) ? 1 : 0);
}

/*
* Makes decisions about spawning or killing threads based on parameters.
*/

void httpcP_handle_load(httpc_t* t, int visits_before, int visits_now,
                                    int clientcount, int totalhandled) {
  int donotreduce = 0;
  if (is_high(visits_before, t->watch.sleeptime) ||
      is_high(visits_now, t->watch.sleeptime)) {
    donotreduce = 1;
  }

  // set new sleeptime based on absolute load
  if (clientcount == 0 || (clientcount > 0 && totalhandled/clientcount <= 1)) {
    if (!donotreduce) httpc_examine_handlers(t, t->watch.minthreads+1);
    t->watch.sleeptime = 3;
  } else {
    int i, wantedhandlers = totalhandled + 1, tobespawned;
    if (wantedhandlers > t->watch.maxthreads) {
      wantedhandlers = t->watch.maxthreads;
    }
    tobespawned = wantedhandlers - t->handlers.count;
    if (tobespawned > 0) {
      fprintf(stderr, "* Spawning %d new handlers.\n", tobespawned);
      for (i=0; i<tobespawned; i++) {
        httpc_spawn_handler(t);
      }
    }
    t->watch.sleeptime = 1;
  }
}

/*
* Creates a snapshot of the current load.
*/

void httpcP_watch_load(httpc_t* t) {
  int v0 = t->watch.visits_1;
  int v1 = t->watch.visits;
  int i, to = 0, cn = t->waitcount;
  t->watch.visits_1 = t->watch.visits;
  t->watch.visits = 0;
  for (i=0; i<t->servers.count; i++) {
    to += t->servers.list[i]->clients.count;
  }
  httpcP_handle_load(t, v0, v1, cn, to);
}

/*
* Removes trailing connections from handlers.
*/

void httpcP_watch_trailers(httpc_t* t) {
  time_t tm = time(NULL);
  int i;
  for (i=0; i<t->servers.count; i++) {
    if (pthread_mutex_lock(&(t->servers.list[i]->clients.mutex)) == 0) {
      int j;
      for (j=0; j<t->servers.list[i]->clients.count; j++) {
        client_t* client = t->servers.list[i]->clients.list[j];
        if (client->handler &&
            client->handler->timestamp < tm - 30 &&
            client->fd >= 0) {
          httpc_close_client_safe(client);
        }
      }
      pthread_mutex_unlock(&(t->servers.list[i]->clients.mutex));
    }
  }
}

/*
* Will send a "\r\n" to all JAVASCRIPT_CAPABLE sucking clients every
* 30 seconds.
*/

void httpcP_watch_suckers(httpc_t* t) {
  if (t->watch.sleeptime * t->watch.recurrence >= 30) {
    int i;
    for (i=0; i<t->servers.count; i++) {
      if (pthread_mutex_lock(&(t->servers.list[i]->clients.mutex)) == 0) {
        int j;
        for (j=0; j<t->servers.list[i]->clients.count; j++) {
          client_t* client = t->servers.list[i]->clients.list[j];
          if (client->sucking &&
              client->session->browser == JAVASCRIPT_CAPABLE &&
              client->writestate < SELECTABLE) {
            //..
            client->writestate = SELECTABLE;
          }
        }
        pthread_mutex_unlock(&(t->servers.list[i]->clients.mutex));
      }
    }
    t->watch.recurrence = 0;
  }
  t->watch.recurrence++;
}

/*
* Will remove all sessions whose client reference count is zero and timestamp
* is older than 30 seconds.
*/

void httpcP_watch_sessions(httpc_t* t) {
  time_t tm = time(NULL);
  if (pthread_mutex_lock(&(t->sessions.mutex)) == 0) {
    int k;
    for (k=0; k<t->sessions.count; k++) {
      if (t->sessions.list[k]->clientrefcount == 0 &&
          t->sessions.list[k]->timestamp < tm - 30) {
#ifdef _DEBUG
        fprintf(stderr, "Session '%s' timed out.\n",
                        t->sessions.list[k]->cookie);
#endif
        httpc_remove_session_safe(t, t->sessions.list[k]);
      }
    }
    pthread_mutex_unlock(&(t->sessions.mutex));
  }
}

void* httpcP_watcherthread(void* arg) {
  httpc_t* t = (httpc_t*)arg;
  while (t->state != APPLICATION_END) {
    httpcP_watch_load(t);
    httpcP_watch_suckers(t);
    httpcP_watch_trailers(t);
    httpcP_watch_sessions(t);
    sleep(t->watch.sleeptime);
  }
  return NULL;
}

int httpc_watcherthread(httpc_t* t) {
  int i, n;
  memset(&(t->watch), 0, sizeof(t->watch));
  if ((n = c_first_child_named(t->config, -1, "maxthreads")) != -1) {
    t->watch.maxthreads = c_value_int(t->config, n);
  }
  if (!t->watch.maxthreads) {
    t->watch.maxthreads = HTTPC_MAXTHREADS;
  }
  if ((n = c_first_child_named(t->config, -1, "minthreads")) != -1) {
    t->watch.minthreads = c_value_int(t->config, n);
  }
  if (!t->watch.minthreads) {
    t->watch.minthreads = HTTPC_MINTHREADS;
  }
  t->watch.sleeptime = 3;
  fprintf(stderr, "* Initially spawning %d handlers.\n", t->watch.minthreads);
  for (i=0; i<t->watch.minthreads; i++) {
    httpc_spawn_handler(t);
  }
  if (pthread_create(&(t->watch.thread), NULL,
                     httpcP_watcherthread, (void*)t)) {
    return -1;
  }
  return 0;
}

/*
* Is called after every time select() in the main thread falls through
* with an inet connection available for reading.
* Used to keep track of load at a given time.
*/

void httpc_touch(httpc_t* t) {
  t->watch.visits++;
}
